==> processing-0.40.zip: processing-0.40/lib/process.py <==
#
# Module providing the `Process` class which emulates `threading.Thread`
#
# processing/process.py
#
# Copyright (c) 2006, 2007, R Oudkerk --- see COPYING.txt
#

__all__ = [
    'Process', 'currentProcess', 'activeChildren', 'freezeSupport',
    'ProcessExit'
    ]

#
# Imports
#

import os
import sys
import signal
import subprocess
import time
import atexit
import weakref
import itertools
import copy_reg
import cPickle

import _processing                              # from . import _processing
from finalize import Finalize, _run_finalizers  # from .finalize import ...
from logger import info                         # form .logger import info

#
# `ProcessExit` exception
#

class ProcessExit(SystemExit):
    pass

#
# Public functions
#

def currentProcess():
    '''
    Return process object representing the current process
    '''
    return _current_process

def activeChildren():
    '''
    Return list of process objects corresponding to live child processes
    '''
    _cleanup()
    return list(_current_process._children)

def freezeSupport():
    '''
    Check whether this is a fake forked process in a frozen executable.
    If so then run code specified by commandline and exit.
    '''
    if sys.platform == 'win32' and getattr(sys, 'frozen', False):
        from processing import _nonforking
        _nonforking.freezeSupport()

#
# The `Process` class
#

class Process(object):
    '''
    Process objects represent activity that is run in a separate process

    The class is analagous to `threading.Thread`
    '''
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        _current_process._counter += 1
        counter = _current_process._counter

        self._target = target
        self._args = tuple(args)
        self._kwargs = kwargs.copy()
        self._stoppable = False
        self._parent_pid = os.getpid()
        self._popen = None
        self._identity = _current_process._identity + (counter,)
        self._authkey = _current_process._authkey
        self._daemonic = _current_process._daemonic
        self._logargs = _current_process._logargs
        self._name = name or 'Process-' + ':'.join(map(str, self._identity))
        self._counter = 0
        self._children = set()

    def run(self):
        '''
        Method to be run in sub-process; can be overridden in sub-class
        '''
        if self._target:
            self._target(*self._args, **self._kwargs)
            
    def start(self):
        '''
        Start child process
        '''
        assert self._popen is None, 'cannot start a process twice'
        assert self._parent_pid == os.getpid(), \
               'can only start a process object created by current process'
        _cleanup()
        sys.stdout.flush()
        sys.stderr.flush()
        self._popen = Popen(self, self._stoppable)
        _current_process._children.add(self)

    def stop(self):
        '''
        Raise `KeyboardInterrupt` in the process to stop it
        '''
        assert self._stoppable, '`setStoppable(True)` was not used'
        self._popen.stop()

    def terminate(self):
        '''
        Terminate process; sends `SIGTERM` signal or uses `TerminateProcess()`
        '''
        self._popen.terminate()
        
    def join(self, timeout=None):
        '''
        Wait until child process terminates
        '''
        assert self._parent_pid == os.getpid(), 'can only join a child process'
        assert self._popen is not None, 'can only join a started process'
        if timeout == 0:
            res = self._popen.poll()
        elif timeout is None:
            # avoid using `Popen.wait()` because that is uninterruptible
            res = _sleep_until_neq(self._popen.poll, None, 1e100)
        else:
            res = _sleep_until_neq(self._popen.poll, None, timeout)
        if res is not None:
            _cleanup()

    def isAlive(self):
        '''
        Return whether child process is alive
        '''
        if self is _current_process:
            return True
        assert self._parent_pid == os.getpid(), 'can only test a child process'
        if self._popen is None:
            return False
        self._popen.poll()
        return self._popen.returncode is None

    def getName(self):
        '''
        Return name of process
        '''
        return self._name

    def setName(self, name):
        '''
        Set name of process
        '''
        assert type(name) is str, 'name must be a string'
        self._name = name

    def isDaemon(self):
        '''
        Return whether process is a daemon
        '''
        return self._daemonic

    def setDaemon(self, daemonic):
        '''
        Set whether process is a daemon
        '''
        assert self._popen is None, 'process has already started'
        assert hasattr(self, 'stop'), 'process needs a `stop()` method'
        self._daemonic = daemonic

    def getAuthKey(self):
        '''
        Return authorization key of process
        '''
        return self._authkey

    def setAuthKey(self, authkey):
        '''
        Set authorization key of process
        '''
        assert type(authkey) is str, 'value must be a string'

        self._authkey = authkey

    def getExitCode(self):
        '''
        Return exit code of process or `None` if it has yet to stop
        '''
        if self._popen is None:
            return self._popen
        if self._popen.poll() == 0x10000:
            return -signal.SIGTERM
        return self._popen.poll()

    def getPid(self):
        '''
        Return PID of process or `None` if it has yet to start
        '''
        return self._popen and self._popen.pid

    def getStoppable(self):
        '''
        Returns whether process supports the `stop()` method and `ProcessExit`
        '''
        return self._stoppable

    def setStoppable(self, value):
        '''
        Set whether process supports the `stop()` method and `ProcessExit`
        '''
        assert self._popen is None, 'process has already started'
        self._stoppable = value

    def __repr__(self):
        if self is _current_process:
            status = 'started'
        elif self._parent_pid != os.getpid():
            status = 'unknown'
        elif self._popen is None:
            status = 'initial'
        else:
            if self._popen.poll() is not None:
                status = self.getExitCode()
            else:
                status = 'started'

        if type(status) is int:
            if status == 0:
                status = 'stopped'
            else:
                status = 'stopped[%s]' % _exitcode_to_name.get(status, status)

        return '<%s(%s, %s%s)>' % (type(self).__name__, self._name,
                                   status, self._daemonic and ' daemon' or '')

    ##

    def _bootstrap(self):
        global _current_process
        try:
            sys.stdin.close()
            Finalize._registry.clear()
            if sys.platform == 'win32' and self._logargs is not None:
                from processing.logger import enableLogging
                enableLogging(*self._logargs)
            _current_process = self
            _afterfork()
            try:
                info('child process calling self.run()')
                self.run()
                exitcode = 0
            finally:
                sys.stdout.flush()
                sys.stderr.flush()
                _exit_func()
        except ProcessExit:
            exitcode = 127
        except SystemExit, e:
            if not e.args:
                exitcode = 1
            elif type(e.args[0]) is int:
                exitcode = e.args[0]
            else:
                print >>sys.stderr, e.args[0]
                exitcode = 1
        except:
            exitcode = 1
            import traceback
            print >>sys.stderr, 'Process %s:' % self.getName()
            traceback.print_exc()

        info('process exiting with `os.exit(...)`')
        os._exit(exitcode)

#
# Create object representing the main process
#

class _MainProcess(Process):

    def __init__(self):
        self._identity = ()
        self._daemonic = False
        self._name = 'MainProcess'
        self._parent_pid = None
        self._popen = None
        self._logargs = None
        self._counter = 0
        self._children = set()

        # calculate authentication key
        try:
            self._authkey = os.urandom(16).encode('hex')
        except AttributeError:
            import random
            key = [chr(random.randrange(256)) for i in range(16)]
            self._authkey = ''.join(key).encode('hex')

_current_process = _MainProcess()
del _MainProcess

#
# Private functions
#

def _cleanup():
    '''
    Purge `_children` of dead processes
    '''
    for p in list(_current_process._children):
        if p._popen.poll() is not None:
            _current_process._children.discard(p)

def _sleep_until_neq(func, value, timeout):
    '''
    Sleep until `func() != value` or timeout elapses
    Returns the last value returned by `func()`
    '''
    deadline = time.time() + timeout
    delay = 0.0005
    while 1:
        res = func()
        if res != value:
            break
        remaining = deadline - time.time()
        if remaining <= 0:
            break
        delay = min(delay * 2, remaining, 0.05)
        time.sleep(delay)
    return res

#
# We define a subclass of `subprocess.Popen`.
# The constructor takes a process object as its argument.
#

if sys.platform != 'win32':

    class Popen(subprocess.Popen):

        def __init__(self, process_obj, ignore=None):
            import random
            subprocess._cleanup()
            self.returncode = None
            self.pid = os.fork()
            if self.pid == 0:
                random.seed()
                process_obj._bootstrap()
                os._exit(0)
            if not hasattr(subprocess.Popen, '__del__'):  # Python 2.4
                subprocess._active.append(self)
            else:                                         # Python 2.5
                self._child_created = True

        def stop(self):
            if self.returncode is None:
                try:
                    os.kill(self.pid, STOP_PROCESS_SIGNAL)
                except OSError:
                    if self.returncode is not None:
                        raise

        def terminate(self):
            if self.returncode is None:
                try:
                    os.kill(self.pid, signal.SIGTERM)
                except OSError:
                    if self.returncode is not None:
                        raise

else:
    
    CREATE_NEW_PROCESS_GROUP = 0x00000200
    
    class Popen(subprocess.Popen):

        def __init__(self, process_obj, new_console=1):
            from processing import _nonforking
            subprocess.Popen.__init__(
                self, _nonforking.get_commandline(),
                stdin=subprocess.PIPE,
                creationflags=new_console and CREATE_NEW_PROCESS_GROUP
                )
            prep_data = _nonforking.get_preparation_data(
                process_obj._name, new_console
                )
            self.stdin.write(cPickle.dumps(prep_data, 2).encode('hex') + '\n')
            self.stdin.write(cPickle.dumps(process_obj, 2).encode('hex'))
            self.stdin.close()
            
        def stop(self):
            if self.returncode is None:
                try:
                    _processing.GenerateConsoleCtrlEvent(1, self.pid)
                except WindowsError:
                    if self.returncode is not None:
                        raise

        def terminate(self):
            if self.returncode is None:
                try:
                    _processing.TerminateProcess(int(self._handle), 0x10000)
                except WindowsError:
                    if self.returncode is not None:
                        raise

#
# In Python 2.5 `subprocess.Popen` has a `__del__()` method which often
# seems to raise `AttributeError` when the interpreter shuts down.  We
# overwrite it.
#

if hasattr(Popen, '__del__'):
    def __del__(self):
        if Popen is not None:
            try:
                super(Popen, self).__del__()
            except AttributeError:
                pass
    Popen.__del__ = __del__
    del __del__

#
# Give names to some return codes
#

_exitcode_to_name = { 127 : 'ProcessExit' }

for name, signum in signal.__dict__.items():
    if name[:3]=='SIG' and '_' not in name:
        _exitcode_to_name[-signum] = name

#
# Signal handling
#

if sys.platform == 'win32':
    STOP_PROCESS_SIGNAL = signal.SIGBREAK
else:
    STOP_PROCESS_SIGNAL = signal.SIGUSR1

    def _STOP_PROCESS_handler(signum, frame):
        raise ProcessExit

    # On Windows signal handler is set in `_nonforking`
    signal.signal(STOP_PROCESS_SIGNAL, _STOP_PROCESS_handler)

#
# Make bound and unbound instance methods and class methods picklable
#

_MethodType = type(Process.start)

def _reduce_method(m):
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.func_name)
    else:
        return getattr, (m.im_self, m.im_func.func_name)

copy_reg.pickle(_MethodType, _reduce_method)

#
# Support for reinitialization of objects when bootstrapping a child process
#

_afterfork_registry = weakref.WeakValueDictionary()

def _afterfork():
    for (ident, func), obj in _afterfork_registry.items():
        func(obj)
        
def _register_afterfork(obj, func):
    _afterfork_registry[(id(obj), func)] = obj
    
#
# Clean up on exit
#

def _exit_func():
    info('running all "atexit" finalizers with priority >= 0')
    _run_finalizers(0)

    for p in activeChildren():
        if p._daemonic and p._stoppable:
            info('calling `stop()` for daemon %s', p.getName())
            p._popen.stop()

    deadline = time.time() + 0.1

    for p in activeChildren():
        if p._daemonic and p._stoppable:
            info('calling `join(timeout)` for daemon %s', p.getName())
            p.join(deadline - time.time())

    for p in activeChildren():
        if p._daemonic:
            info('calling `terminate()` for daemon %s', p.getName())
            p._popen.terminate()

    for p in activeChildren():
        info('calling `join()` for process %s', p.getName())
        p.join()

    info('running all "atexit" finalizers with priority < 0')
    _run_finalizers()

atexit.register(_exit_func)

==> processing-0.50.zip: processing-0.50/lib/process.py <==
#
# Module providing the `Process` class which emulates `threading.Thread`
#
# processing/process.py
#
# Copyright (c) 2006-2008, R Oudkerk --- see COPYING.txt
#

__all__ = [
    'Process', 'currentProcess', 'activeChildren'
    ]

#
# Imports
#

import os
import sys
import time
import signal
import atexit
import weakref
import copy_reg
import itertools

#
# Public functions
#

def currentProcess():
    '''
    Return process object representing the current process
    '''
    return _current_process

def activeChildren():
    '''
    Return list of process objects corresponding to live child processes
    '''
    _cleanup()
    return list(_current_process._children)

def _cleanup():
    '''
    Purge `_children` of dead processes
    '''
    for p in list(_current_process._children):
        if p._popen.poll() is not None:
            _current_process._children.discard(p)

#
# The `Process` class
#

class Process(object):
    '''
    Process objects represent activity that is run in a separate process

    The class is analagous to `threading.Thread`
    '''
    def __init__(self, group=None, target=None, name=None, args=(), kwargs={}):
        counter = _current_process._counter.next()

        self._identity = _current_process._identity + (counter,)
        self._authkey = _current_process._authkey
        self._daemonic = _current_process._daemonic
        self._logargs = _current_process._logargs
        self._parent_pid = os.getpid()
        self._popen = None

        self._target = target
        self._args = tuple(args)
        self._kwargs = kwargs.copy()
        self._name = name or 'Process-' + ':'.join(map(str, self._identity))

    def run(self):
        '''
        Method to be run in sub-process; can be overridden in sub-class
        '''
        if self._target:
            self._target(*self._args, **self._kwargs)
            
    def start(self):
        '''
        Start child process
        '''
        from processing.forking import Popen
        assert self._popen is None, 'cannot start a process twice'
        assert self._parent_pid == os.getpid(), \
               'can only start a process object created by current process'
        _cleanup()
        self._popen = Popen(self)
        _current_process._children.add(self)

    def terminate(self):
        '''
        Terminate process; sends `SIGTERM` signal or uses `TerminateProcess()`
        '''
        self._popen.terminate()
        
    def join(self, timeout=None):
        '''
        Wait until child process terminates
        '''
        assert self._parent_pid == os.getpid(), 'can only join a child process'
        assert self._popen is not None, 'can only join a started process'
        if timeout == 0:
            res = self._popen.poll()
        elif timeout is None:
            res = self._popen.wait()
        else:
            res = self._popen.wait_timeout(timeout)
        if res is not None:
            _cleanup()

    def isAlive(self):
        '''
        Return whether child process is alive
        '''
        if self is _current_process:
            return True
        assert self._parent_pid == os.getpid(), 'can only test a child process'
        if self._popen is None:
            return False
        self._popen.poll()
        return self._popen.returncode is None

    def getName(self):
        '''
        Return name of process
        '''
        return self._name

    def setName(self, name):
        '''
        Set name of process
        '''
        assert type(name) is str, 'name must be a string'
        self._name = name

    def isDaemon(self):
        '''
        Return whether process is a daemon
        '''
        return self._daemonic

    def setDaemon(self, daemonic):
        '''
        Set whether process is a daemon
        '''
        assert self._popen is None, 'process has already started'
        self._daemonic = daemonic

    def getAuthKey(self):
        '''
        Return authorization key of process
        '''
        return self._authkey

    def setAuthKey(self, authkey):
        '''
        Set authorization key of process
        '''
        assert type(authkey) is str, 'value must be a string'
        self._authkey = authkey

    def getExitCode(self):
        '''
        Return exit code of process or `None` if it has yet to stop
        '''
        if self._popen is None:
            return self._popen
        return self._popen.poll()

    def getPid(self):
        '''
        Return PID of process or `None` if it has yet to start
        '''
        if self is _current_process:
            return os.getpid()
        else:
            assert self._parent_pid == os.getpid()
            return self._popen and self._popen.pid

    def __repr__(self):
        if self is _current_process:
            status = 'started'
        elif self._parent_pid != os.getpid():
            status = 'unknown'
        elif self._popen is None:
            status = 'initial'
        else:
            if self._popen.poll() is not None:
                status = self.getExitCode()
            else:
                status = 'started'

        if type(status) is int:
            if status == 0:
                status = 'stopped'
            else:
                status = 'stopped[%s]' % _exitcode_to_name.get(status, status)

        return '<%s(%s, %s%s)>' % (type(self).__name__, self._name,
                                   status, self._daemonic and ' daemon' or '')

    ##

    def _bootstrap(self):
        from processing.finalize import _registry
        from processing.logger import info
        
        global _current_process
        try:
            self._children = set()
            self._counter = itertools.count(1)
            sys.stdin.close()
            _registry.clear()
            if sys.platform == 'win32' and self._logargs is not None:
                from processing.logger import enableLogging
                enableLogging(*self._logargs)
            _current_process = self
            _afterfork()
            info('child process calling self.run()')
            try:
                self.run()
                exitcode = 0
            finally:
                _exit_func()
        except SystemExit, e:
            if not e.args:
                exitcode = 1
            elif type(e.args[0]) is int:
                exitcode = e.args[0]
            else:
                print >>sys.stderr, e.args[0]
                exitcode = 1
        except:
            exitcode = 1
            import traceback
            print >>sys.stderr, 'Process %s:' % self.getName()
            traceback.print_exc()

        info('process exiting with exitcode %d' % exitcode)
        return exitcode

#
# Create object representing the main process
#

class _MainProcess(Process):

    def __init__(self):
        self._identity = ()
        self._daemonic = False
        self._name = 'MainProcess'
        self._parent_pid = None
        self._popen = None
        self._logargs = None
        self._counter = itertools.count(1)
        self._children = set()
        self._authkey = ''.join('%02x' % ord(c) for c in os.urandom(16))


_current_process = _MainProcess()
del _MainProcess

#
# Give names to some return codes
#

_exitcode_to_name = {}

for name, signum in signal.__dict__.items():
    if name[:3]=='SIG' and '_' not in name:
        _exitcode_to_name[-signum] = name

#
# Make bound and unbound instance methods and class methods picklable
#

def _reduce_method(m):
    if m.im_self is None:
        return getattr, (m.im_class, m.im_func.func_name)
    else:
        return getattr, (m.im_self, m.im_func.func_name)

copy_reg.pickle(type(_current_process.start), _reduce_method)

#
# Support for reinitialization of objects when bootstrapping a child process
#

_afterfork_registry = weakref.WeakValueDictionary()
_afterfork_count = itertools.count()

def _afterfork():
    # execute in order of registration
    for (index, ident, func), obj in sorted(_afterfork_registry.items()):
        func(obj)
        
def _register_afterfork(obj, func):
    _afterfork_registry[(_afterfork_count.next(), id(obj), func)] = obj
    
#
# Clean up on exit
#

def _exit_func():
    from processing.finalize import _run_finalizers
    from processing.logger import info
    
    _current_process._exiting = True

    info('running all "atexit" finalizers with priority >= 0')
    _run_finalizers(0)

    for p in activeChildren():
        if p._daemonic:
            info('calling `terminate()` for daemon %s', p.getName())
            p._popen.terminate()

    for p in activeChildren():
        info('calling `join()` for process %s', p.getName())
        p.join()

    info('running the remaining "atexit" finalizers')
    _run_finalizers()

atexit.register(_exit_func)
